unit MCIStreamPlayer;

interface

uses
  Windows, SysUtils, Classes, ExtCtrls;

type
  TStreamReadEvent=procedure(ASender:TObject; ACount:integer) of object;
  TStreamSeekEvent=procedure(ASender:TObject; AOffset:integer) of object;

  TMCIStreamPlayer = class(TPanel)
  private
    FIsOpened: boolean;
    FDeviceID: Word;
    FDeviceName:string;
//    FFileName: string;
    FOpened: boolean;
    FAutoPlay: boolean;
    FStream:TStream;
    FStreamOriginalPosition:int64;
    FOnStreamClose: TNotifyEvent;
    FOnStreamOpen: TNotifyEvent;
    FOnStreamRead: TStreamReadEvent;
    FOnStreamSeek: TStreamSeekEvent;

//    procedure SetFileName(const Value: string);
    procedure DoOpen;
    procedure DoClose;
    procedure DoSeek(lParam1, lParam2:LPARAM; var offset:longint);
    function DoRead(var buffer; count:integer):integer;
  public
    property Stream:TStream read FStream write FStream;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure open;
    procedure close;
  published
//    property FileName:string read FFileName write SetFileName;
    property AutoPlay:boolean read FAutoPlay write FAutoPlay default true;
    property Opened:boolean read FOpened;

    property OnStreamRead:TStreamReadEvent read FOnStreamRead write FOnStreamRead;
    property OnStreamOpen:TNotifyEvent read FOnStreamOpen write FOnStreamOpen;
    property OnStreamClose:TNotifyEvent read FOnStreamClose write FOnStreamClose;
    property OnStreamSeek:TStreamSeekEvent read FOnStreamSeek write FOnStreamSeek;
  end;

  EMCIStreamPlayerException=class(Exception);

procedure Register;

implementation

uses MMSystem;

const
  FFourCC = 'IOCB'; // acronym for IO CallBack (but it really does not matter - choose as you wish)

resourcestring
  CannotInstallIOProc = 'Could not install IO proc.';
  FileNotFound = 'File to be played "%s" was not found.';
  OpenError = 'Some error occured while opening: %s';
  NotOpen = 'The device could not be opened.';
  OnlyOneInstance = 'Only one player instance is supported at this moment.';
  WindowNotSet = 'Playing window is not set.';
  StreamNotAssigned = 'Stream not assigned';

procedure Register;
begin
  RegisterComponents('Ciuly', [TMCIStreamPlayer]);
end;

var mciPlayer:TMCIStreamPlayer=nil;

function MyIOCallback(lpmmioinfo: PChar; uMessage: UINT; lParam1, lParam2: LPARAM): Longint stdcall;
// This function is called and will provide access to the file. The MM system "thinks"
// this is a file, while we use the global memory stream.
var
  AInfo: pmmioInfo; c:integer;
begin
  AInfo := pmmioInfo(lpmmioinfo);
  case uMessage of
  MMIOM_OPEN:
    // Open the "file"
    begin
      if mciPlayer.FOpened then
      begin
        Result := 0;
        exit;
      end;
      mciPlayer.DoOpen;
//      FileMode:=fmOpenRead;// reading from a readonly running exe
      AInfo.lDiskOffset := 0;
//      f.seek(startoffset,soFromBeginning);// reading from an exe, where the movie is at the end. need to know the startoffset
      Result := 0;
    end;
  MMIOM_CLOSE:
      // Close the "file"
    begin
      mciPlayer.DoClose;
      Result := 0;
    end;
  MMIOM_READ:
    // Read data from the "file"
    begin
      // Copy from memory - no checking
{      c:=f.read(pointer(lParam1)^,lParam2);
      AInfo.lDiskOffset := AInfo.lDiskOffset + lParam2;
      Result := lParam2;{}
      c:=mciPlayer.DoRead(pointer(lParam1)^,lParam2);
      AInfo.lDiskOffset := AInfo.lDiskOffset + c;
      Result := c;
    end;
  MMIOM_SEEK:
    // Seek new position in the "file"
    begin
//      f.seek(startoffset+AInfo.lDiskOffset,soFromBeginning);// see above about the startoffset
      mciPlayer.DoSeek(lParam1, lParam2, AInfo.lDiskOffset);
      Result := AInfo.lDiskOffset;
    end;
  else
    // Unexpected msgs. For instance, we do not process MMIOM_WRITE in this sample
    Result := -1;
  end;
end;

function findStr(subStr,str:string; start:integer):integer;
var s:string;
begin
  s:=copy(str,start,length(str));
  result:=start+pos(subStr,s)-1;
end;

{ TMCIStreamPlayer }

procedure TMCIStreamPlayer.DoClose;
begin
  if FStream<>nil then
  begin
//    if FileExists(FFileName) then
//      FreeAndNil(FStream);
  end;            
  FOpened:=false;
end;

constructor TMCIStreamPlayer.Create(AOwner: TComponent);
begin
  inherited;

  if mciPlayer<>nil then
    raise EMCIStreamPlayerException.Create(OnlyOneInstance);
  mciPlayer:=self;
    
  FOpened:=False;
  FAutoPlay:=true;
end;

destructor TMCIStreamPlayer.Destroy;
begin
  close;
  inherited;
end;

procedure TMCIStreamPlayer.DoOpen;
begin
  if FStream=nil then
  begin
{    if FileExists(FFileName) then
    begin
      FOpened:=true;
      FStream:=TFileStream.Create(filename,fmOpenRead or fmShareDenyNone)
    end                      else
      raise EMCIStreamPlayerException.Create(format(FileNotFound, [FFileName]));{}
    raise EMCIStreamPlayerException.Create(StreamNotAssigned);  
  end            else
  begin
    FStream.Position:=FStreamOriginalPosition;
    FOpened:=true;
  end;
end;

function TMCIStreamPlayer.DoRead(var buffer; count: integer): integer;
begin
  result:=FStream.Read(buffer, count);
end;

procedure TMCIStreamPlayer.DoSeek(lParam1, lParam2: LPARAM; var offset: longint);
begin
  case lParam2 of
    // From beginning:
    0: Offset := lParam1;
    // From current:
    1: Offset := Offset + lParam1;
    // From end:
    2: Offset := FStream.size - 1 - lParam1;
  end;
  FStream.seek(offset, soFromBeginning);
end;

{procedure TMCIStreamPlayer.SetFileName(const Value: string);
begin
  if not FileExists(Value) then
    raise EMCIStreamPlayerException.Create(format(FileNotFound, [Value]));
  FFileName:=Value;
end;{}

procedure TMCIStreamPlayer.open;
var
  OpenParm: TMCI_Open_Parms;
  GenParm: TMCI_Generic_Parms;
  FFlags: Longint;
  FError: Longint;
  ErrMsg: array[0..4095] of Char;
  AError: string;

  PlayParm: TMCI_Play_Parms;
  AWindowParm: TMCI_Anim_Window_Parms;
  StatusParms: TMCI_STATUS_PARMS;
  SetParams: TMCI_Set_Parms;

  size:PAnsiChar;
  s1,s2:integer;
  w,h:integer;
begin
{  if (FStream=nil) and (not FileExists(FileName)) then
    raise EMCIStreamPlayerException.Create(format(FileNotFound, [FileName]));{}
  if FStream=nil then
    raise EMCIStreamPlayerException.Create(StreamNotAssigned);
  FStreamOriginalPosition:=FStream.Position;

  if not assigned(mmioInstallIOProc(mmioStringToFourCC(FFourCC, 0), @MyIOCallback, MMIO_INSTALLPROC or MMIO_GLOBALPROC)) then
    raise EMCIStreamPlayerException.Create(CannotInstallIOProc);

  // Open the file
  if FIsOpened then
  begin
    GenParm.dwCallback := Handle;
    mciSendCommand(FDeviceID, mci_Close, 0, Longint(@GenParm));
    FIsOpened := False;
  end;

  FillChar(OpenParm, SizeOf(TMCI_Open_Parms), 0);

  // Open a specific device type - obviously, and unfortunately,
  // the memory play function does NOT work for mpegvideo.
//        OpenParm.lpstrDeviceType := 'MPEGVideo'; // does not work
  OpenParm.lpstrDeviceType := 'avivideo';

  // Set the name of the file with extension .IOCB+ to invoke our own
  // I/O callbacks. This way the file is read from memory
  OpenParm.lpstrElementName :=  PAnsiChar('bogus.'+FFourCC+'+');// TODO: not sure if this should be .IOCB or the FourCC
  // OpenParm.lpstrElementName := PChar(edFilename.Text); // alternative, directly read file
  OpenParm.dwCallback := Handle;
  OpenParm.lpstrAlias:=PAnsiChar('movie'+inttostr(random(100)));

  FFlags := mci_Open_Type or MCI_OPEN_ELEMENT or MCI_OPEN_ALIAS;
  FError := mciSendCommand(0, mci_Open, FFlags, Longint(@OpenParm));

  if FError <> 0 then
  begin
    // problem opening device
    if not mciGetErrorString(FError, ErrMsg, SizeOf(ErrMsg)) then
      AError := 'Unknown error'
    else
      SetString(AError, ErrMsg, StrLen(ErrMsg));
    raise EMCIStreamPlayerException.Create(format(OpenError, [ErrMsg]));
  end else
  begin
    FIsOpened := True;
    FDeviceID := OpenParm.wDeviceID;
    FDeviceName:= OpenParm.lpstrAlias;

    getmem(size,129);
    try
//          mciSendCommand(FDeviceID,mci_
      if mciSendString(PAnsiChar('Where ' + OpenParm.lpstrAlias + ' destination'), size, 128, 0)=0 then
      begin
        s1:=findStr(' ',size,1);
        s2:=findStr(' ',size,s1 + 1);
        s1:=findStr(' ',size,s2 + 1);
        w:=strtoint(copy(size, s2, s1 - s2));
        h:=strtoint(copy(size, s1 + 1,length(size)));
        if (w>0) and (h>0) then
        begin
          ClientWidth:=w;
          ClientHeight:=h;
{          left:=0;// video window position
          top:=0;{}
          Caption:='Video Window Caption';
//          show;
        end;
      end;
    finally
      freemem(size);
    end;
//   start playing
    if not FIsOpened then
      raise EMCIStreamPlayerException.Create(NotOpen);
    AWindowParm.Wnd := Longint({FWindow.{}Handle);
//      FError :=
    mciSendCommand(FDeviceID, mci_Window, mci_Wait or mci_Anim_Window_hWnd, Longint(@AWindowParm) );

    SetParams.dwTimeFormat:=MCI_FORMAT_MILLISECONDS;
    mciSendCommand(FDeviceID,MCI_SET,MCI_SET_TIME_FORMAT,Longint(@SetParams));

    StatusParms.dwItem:=MCI_STATUS_LENGTH;
    StatusParms.dwTrack:=0;
    mciSendCommand(FDeviceID,MCI_STATUS,MCI_STATUS_ITEM,longint(@StatusParms));
//    total:=StatusParms.dwReturn;
//    TrackBar.Max:=total div 1000;
//    debug('MCI - total length (in seconds) = '+inttostr(TrackBar.Max));

//    totalLength:=total div 1000;

      // play
    PlayParm.dwCallback := Handle;
//      FError :=
    if AutoPlay then
      mciSendCommand(FDeviceID, mci_Play, 0, Longint(@PlayParm));
  end;
end;

procedure TMCIStreamPlayer.Close;
var
  GenParm: TMCI_Generic_Parms;
begin
  if FIsOpened then
  begin
//    GenParm.dwCallback := Handle;
    mciSendCommand(FDeviceID, mci_Close, 0, Longint(@GenParm));
    FIsOpened := False;
    // remove mmio callback
    mmioInstallIOProc(mmioStringToFourCC(FFourCC, 0), nil, MMIO_REMOVEPROC);
  end;
  FreeAndNil(FStream);
end;

initialization
  randomize;

end.
 